{% extends "!layout.html" %}

{% block relbar1 %}
<div style=" background-image:url('{{
pathto("_static/bg.png", 1) }}'); text-align: left; padding: 0px 0px 0px 0px">
<a href="{{ pathto('index') }}"><img src="{{
pathto("_static/header.png", 1) }}" border="0" alt="masai"/></a>
</div>
{{ super() }}
{% endblock %}

{# put the sidebar before the body #}
{% block sidebar1 %}{{ sidebar() }}{% endblock %}
{% block sidebar2 %}{% endblock %}



