@import "sphinxdoc.css";

/**
 * 	Separation between classes and functions
 */

dl.class, dl.function {
    border-top: 2px solid #888;
}

dl.method, dl.attribute {
    border-top: 1px solid #aaa;
}

dl.glossary dt {
    font-weight: bold;
    font-size: 1.1em;
}

/*
 * Title
 */
h2 {
  border-bottom:2px solid #ddd;
  font-size:1.35em;
  margin:1.3em 0 0.2em;
  padding:0;
}

th {
  background-color:#EDEFED;
}

table.field-list th {
  border-left-color:#AAAAAA !important;
  border-left-style:solid !important;
  border-left-width:1px !important;
  padding-left:5px;
}