#!/usr/bin/python
# -*- coding: utf-8 -*-
#===============================================================================
# __init__
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================

"""The **MASAÏ** package is intended to provide two main services:  

1. an API, which can be used in other user applications

2. a GUI, providing extensible tools for processing, analysing and fitting of 
   solid state NMR spectra. 

.. note:: 

    In version 0.1, only 1D processing and fitting tools are provided

.. seealso:: the :ref:`installation` guide for install instructions


"""